@@ -1,7 +1,10 @@
-import json
+import base64
+import dataclasses
 import logging
 import threading
-from typing import cast, Mapping, Optional, Generator, List, Union
+from datetime import datetime, timezone
+from time import time
+from typing import cast, Optional, Generator, List, Union, Dict
 from unittest import mock
 
 import boto3
@@ -18,37 +21,34 @@ from hedwig.models import Message
 logger = logging.getLogger(__name__)
 
 
+@dataclasses.dataclass(frozen=True)
 class AWSMetadata:
     """
     AWS specific metadata for a Message
     """
 
-    def __init__(self, receipt: str):
-        self._receipt: str = receipt
+    receipt: str
+    """
+    AWS receipt identifier
+    """
 
-    @property
-    def receipt(self) -> str:
-        """
-        AWS receipt identifier
-        """
-        return self._receipt
+    first_receive_time: datetime
+    """
+    The time the message was first received from the queue. The value
+    is calculated as best effort and is approximate.
+    """
 
-    def __eq__(self, o: object) -> bool:
-        if not isinstance(o, AWSMetadata):
-            return False
-        return self._receipt == o._receipt
+    sent_time: datetime
+    """
+    Time this message was originally sent to AWS
+    """
 
-    def __ne__(self, o: object) -> bool:
-        return not self.__eq__(o)
-
-    def __str__(self) -> str:
-        return repr(self)
-
-    def __repr__(self) -> str:
-        return f'AWSMetadata(receipt={self._receipt})'
-
-    def __hash__(self) -> int:
-        return hash((self._receipt,))
+    receive_count: int
+    """
+    The receive count received from SQS.
+    The first delivery of a given message will have this value as 1. The value
+    is calculated as best effort and is approximate.
+    """
 
 
 class AWSSNSPublisherBackend(HedwigPublisherBaseBackend):
@@ -70,26 +70,50 @@ class AWSSNSPublisherBackend(HedwigPublisherBaseBackend):
             )
         return self._sns_client
 
-    @staticmethod
-    def _get_sns_topic(message: Message) -> str:
-        return f'arn:aws:sns:{settings.AWS_REGION}:{settings.AWS_ACCOUNT_ID}:hedwig-{message.topic}'
+    @classmethod
+    def _get_sns_topic(cls, message: Message) -> str:
+        topic = cls.topic(message)
+        if isinstance(topic, tuple):
+            topic, account_id = topic
+        else:
+            account_id = settings.AWS_ACCOUNT_ID
+        return f'arn:aws:sns:{settings.AWS_REGION}:{account_id}:hedwig-{topic}'
 
     @retry(stop_max_attempt_number=3, stop_max_delay=3000)
-    def _publish_over_sns(self, topic: str, message_json: str, message_attributes: Mapping) -> str:
-        # transform (http://boto.cloudhackers.com/en/latest/ref/sns.html#boto.sns.SNSConnection.publish)
-        message_attributes = {k: {'DataType': 'String', 'StringValue': str(v)} for k, v in message_attributes.items()}
-        response = self.sns_client.publish(TopicArn=topic, Message=message_json, MessageAttributes=message_attributes)
-        return response['PublishResponse']['PublishResult']['MessageId']
+    def _publish_over_sns(self, topic: str, message_payload: str, attributes: Dict[str, str]) -> str:
+        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sns.html#SNS.Topic.publish
+        message_attributes = {str(k): {'DataType': 'String', 'StringValue': str(v)} for k, v in attributes.items()}
+        response = self.sns_client.publish(
+            TopicArn=topic, Message=message_payload, MessageAttributes=message_attributes
+        )
+        return response['MessageId']
 
     def _mock_queue_message(self, message: Message) -> mock.Mock:
         sqs_message = mock.Mock()
-        sqs_message.body = json.dumps(message.as_dict())
+        payload, attributes = message.serialize()
+        # SQS requires UTF-8 encoded string
+        if isinstance(payload, bytes):
+            payload = base64.encodebytes(payload).decode()
+            attributes['hedwig_encoding'] = 'base64'
+        sqs_message.body = payload
+        sqs_message.message_attributes = {
+            k: {'DataType': 'String', 'StringValue': str(v)} for k, v in attributes.items()
+        }
+        sqs_message.attributes = {
+            'ApproximateReceiveCount': 1,
+            'SentTimestamp': int(time() * 1000),
+            'ApproximateFirstReceiveTimestamp': int(time() * 1000),
+        }
         sqs_message.receipt_handle = 'test-receipt'
         return sqs_message
 
-    def _publish(self, message: Message, payload: str, headers: Optional[Mapping] = None) -> str:
+    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> str:
         topic = self._get_sns_topic(message)
-        return self._publish_over_sns(topic, payload, headers)
+        # SNS requires UTF-8 encoded string
+        if isinstance(payload, bytes):
+            payload = base64.encodebytes(payload).decode()
+            attributes['hedwig_encoding'] = 'base64'
+        return self._publish_over_sns(topic, payload, attributes)
 
 
 class AWSSQSConsumerBackend(HedwigConsumerBaseBackend):
@@ -142,6 +166,7 @@ class AWSSQSConsumerBackend(HedwigConsumerBaseBackend):
         params = {
             'MaxNumberOfMessages': num_messages,
             'WaitTimeSeconds': self.WAIT_TIME_SECONDS,
+            'AttributeNames': ['All'],
             'MessageAttributeNames': ['All'],
         }
         if visibility_timeout is not None:
@@ -149,9 +174,24 @@ class AWSSQSConsumerBackend(HedwigConsumerBaseBackend):
         return self._get_queue().receive_messages(**params)
 
     def process_message(self, queue_message) -> None:
-        message_json = queue_message.body
+        attributes = {k: o['StringValue'] for k, o in (queue_message.message_attributes or {}).items()}
+        # body is always UTF-8 string
+        message_payload = queue_message.body
+        if attributes.get("hedwig_encoding") == "base64":
+            message_payload = base64.decodebytes(message_payload.encode())
         receipt = queue_message.receipt_handle
-        self.message_handler(message_json, AWSMetadata(receipt))
+        self.message_handler(
+            message_payload,
+            attributes,
+            AWSMetadata(
+                receipt,
+                datetime.fromtimestamp(
+                    int(queue_message.attributes['ApproximateFirstReceiveTimestamp']) / 1000, tz=timezone.utc
+                ),
+                datetime.fromtimestamp(int(queue_message.attributes['SentTimestamp']) / 1000, tz=timezone.utc),
+                int(queue_message.attributes['ApproximateReceiveCount']),
+            ),
+        )
 
     def ack_message(self, queue_message) -> None:
         queue_message.delete()
@@ -246,6 +286,9 @@ class AWSSNSConsumerBackend(HedwigConsumerBaseBackend):
 
     def process_message(self, queue_message) -> None:
         settings.HEDWIG_PRE_PROCESS_HOOK(sns_record=queue_message)
-        message_json = queue_message['Sns']['Message']
-        self.message_handler(message_json, None)
+        message_payload = queue_message['Sns']['Message']
+        attributes = {k: o['Value'] for k, o in queue_message['Sns']['MessageAttributes'].items()}
+        if attributes.get("hedwig_encoding") == "base64":
+            message_payload = base64.decodebytes(message_payload.encode()).decode()
+        self.message_handler(message_payload, attributes, None)
         settings.HEDWIG_POST_PROCESS_HOOK(sns_record=queue_message)
