@@ -1,14 +1,9 @@
-import copy
-import json
 import logging
 import threading
 import uuid
 from concurrent.futures import Future
-from decimal import Decimal
-from typing import Optional, Mapping, Union, Generator, List
-from unittest import mock
+from typing import Optional, Union, Generator, List, Any, Dict, Tuple
 
-from hedwig.backends.import_utils import import_class
 from hedwig.conf import settings
 from hedwig.exceptions import ValidationError, IgnoreException, LoggingException, RetryException
 from hedwig.models import Message
@@ -17,22 +12,16 @@ from hedwig.models import Message
 logger = logging.getLogger(__name__)
 
 
-class HedwigBaseBackend:
+class HedwigPublisherBaseBackend:
     @classmethod
-    def build(cls, dotted_path: str, *args, **kwargs):
+    def topic(cls, message: Message) -> Union[str, Tuple[str, str]]:
         """
-        Import a dotted module path and return the backend class instance.
-        Raise ImportError if the import failed.
+        The topic name for routing the message. When publishing cross-project topics, returned value may be a tuple of
+        topic name and project id for Google or account id for AWS.
         """
-        backend_cls = import_class(dotted_path)
-        return backend_cls(*args, **kwargs)
+        version_pattern = f'{message.major_version}.*'
+        return settings.HEDWIG_MESSAGE_ROUTING[(message.type, version_pattern)]
 
-    @staticmethod
-    def message_payload(data: dict) -> str:
-        return json.dumps(data, default=_decimal_json_default, allow_nan=False)
-
-
-class HedwigPublisherBaseBackend(HedwigBaseBackend):
     def _dispatch_sync(self, message: Message) -> None:
         from hedwig.backends.utils import get_consumer_backend
 
@@ -42,33 +31,47 @@ class HedwigPublisherBaseBackend(HedwigBaseBackend):
         consumer_backend.process_message(queue_message)
         settings.HEDWIG_POST_PROCESS_HOOK(**consumer_backend.post_process_hook_kwargs(queue_message))
 
-    def _mock_queue_message(self, message: Message) -> mock.Mock:
-        return NotImplementedError
+    def _mock_queue_message(self, message: Message):
+        """
+        Generate a mock queue message in proper format as expected by the transport backend. This is primarily used for
+        testing.
+        """
+        raise NotImplementedError
 
-    def _publish(self, message: Message, payload: str, headers: Optional[Mapping] = None) -> Union[str, Future]:
+    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> Union[str, Future]:
+        """
+        Actually publish a message with the given payload and attributes. Some transport mechanisms restrict payload
+        to string formats, and others may require bytes. It's up to the implementing class to encode the payload
+        properly. The implementing class is responsible for decoding the payload back to the same format (bytes or
+        string) as serialized by the validator.
+        """
         raise NotImplementedError
 
     def publish(self, message: Message) -> Union[str, Future]:
+        """
+        Publish a message
+        :return: Either a future if publisher is async, message id otherwise. If a future is returned, it'll result in
+        message id once completed
+        """
         if settings.HEDWIG_SYNC:
             self._dispatch_sync(message)
             return str(uuid.uuid4())
 
-        message_body = message.as_dict()
-        headers = {**settings.HEDWIG_DEFAULT_HEADERS(message=message), **message_body['metadata']['headers']}
-        # make a copy to prevent changing "headers" variable contents in
-        # pre serialize hook
-        message_body['metadata']['headers'] = copy.deepcopy(headers)
-        settings.HEDWIG_PRE_SERIALIZE_HOOK(message_data=message_body)
-        payload = self.message_payload(message_body)
+        default_headers = settings.HEDWIG_DEFAULT_HEADERS(message=message)
+        if default_headers:
+            new_headers = {**default_headers, **message.headers}
+            message = message.with_headers(new_headers)
 
-        result = self._publish(message, payload, headers)
+        payload, attributes = message.serialize()
 
-        log_published_message(message_body, result)
+        result = self._publish(message, payload, attributes)
+
+        log_published_message(message, result)
 
         return result
 
 
-class HedwigConsumerBaseBackend(HedwigBaseBackend):
+class HedwigConsumerBaseBackend:
     @staticmethod
     def pre_process_hook_kwargs(queue_message) -> dict:
         return {}
@@ -77,9 +80,9 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
     def post_process_hook_kwargs(queue_message) -> dict:
         return {}
 
-    def message_handler(self, message_json: str, provider_metadata) -> None:
-        message = self._build_message(message_json, provider_metadata)
-        _log_received_message(message.as_dict())
+    def message_handler(self, message_payload: Union[str, bytes], attributes: dict, provider_metadata) -> None:
+        message = self._build_message(message_payload, attributes, provider_metadata)
+        _log_received_message(message)
 
         message.exec_callback()
 
@@ -97,7 +100,7 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
                     settings.HEDWIG_PRE_PROCESS_HOOK(**self.pre_process_hook_kwargs(queue_message))
                 except Exception:
                     logger.exception(
-                        f'Exception in post process hook for message', extra={'queue_message': queue_message}
+                        'Exception in pre process hook for message', extra={'queue_message': queue_message}
                     )
                     self.nack_message(queue_message)
                     continue
@@ -105,7 +108,7 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
                 try:
                     self.process_message(queue_message)
                 except IgnoreException:
-                    logger.info(f'Ignoring task', extra={'queue_message': queue_message})
+                    logger.info('Ignoring task', extra={'queue_message': queue_message})
                 except LoggingException as e:
                     # log with message and extra
                     logger.exception(str(e), extra=e.extra)
@@ -117,7 +120,7 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
                     self.nack_message(queue_message)
                     continue
                 except Exception:
-                    logger.exception(f'Exception while processing message')
+                    logger.exception('Exception while processing message')
                     self.nack_message(queue_message)
                     continue
 
@@ -125,7 +128,7 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
                     settings.HEDWIG_POST_PROCESS_HOOK(**self.post_process_hook_kwargs(queue_message))
                 except Exception:
                     logger.exception(
-                        f'Exception in post process hook for message', extra={'queue_message': queue_message}
+                        'Exception in post process hook for message', extra={'queue_message': queue_message}
                     )
                     self.nack_message(queue_message)
                     continue
@@ -133,7 +136,7 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
                 try:
                     self.ack_message(queue_message)
                 except Exception:
-                    logger.exception(f'Exception while deleting message', extra={'queue_message': queue_message})
+                    logger.exception('Exception while deleting message', extra={'queue_message': queue_message})
 
     def extend_visibility_timeout(self, visibility_timeout_s: int, metadata) -> None:
         """
@@ -177,32 +180,20 @@ class HedwigConsumerBaseBackend(HedwigBaseBackend):
         raise NotImplementedError
 
     @staticmethod
-    def _build_message(message_json: str, provider_metadata) -> Message:
+    def _build_message(message_payload: Union[str, bytes], attributes: dict, provider_metadata: Any) -> Message:
         try:
-            message_body = json.loads(message_json)
-            settings.HEDWIG_POST_DESERIALIZE_HOOK(message_data=message_body)
-            message = Message(json.loads(message_json))
-            message.validate_callback()
-            message.metadata.provider_metadata = provider_metadata
+            message = Message.deserialize(message_payload, attributes, provider_metadata)
+            # side-effect: validates the callback
+            _ = message.callback
             return message
-        except (ValidationError, ValueError):
-            _log_invalid_message(message_json)
+        except ValidationError:
+            _log_invalid_message(message_payload)
             raise
 
 
-def _decimal_json_default(obj):
-    if isinstance(obj, Decimal):
-        int_val = int(obj)
-        if int_val == obj:
-            return int_val
-        else:
-            return float(obj)
-    raise TypeError
-
-
-def log_published_message(message_body: dict, result: Union[str, Future]) -> None:
+def log_published_message(message: Message, result: Union[str, Future]) -> None:
     def _log(message_id: str):
-        logger.debug('Sent message', extra={'message_body': message_body, 'message_id': message_id})
+        logger.debug('Sent message', extra={'hedwig_message': message, 'message_id': message_id})
 
     if isinstance(result, Future):
         result.add_done_callback(lambda f: _log(f.result()))
@@ -210,9 +201,9 @@ def log_published_message(message_body: dict, result: Union[str, Future]) -> Non
         _log(result)
 
 
-def _log_received_message(message_body: dict) -> None:
-    logger.debug('Received message', extra={'message_body': message_body})
+def _log_received_message(message: Message) -> None:
+    logger.debug('Received message', extra={'hedwig_message': message})
 
 
-def _log_invalid_message(message_json: str) -> None:
-    logger.error('Received invalid message', extra={'message_json': message_json})
+def _log_invalid_message(message_payload: Union[str, bytes]) -> None:
+    logger.error('Received invalid message', extra={'message_payload': message_payload})
