@@ -1,19 +1,20 @@
+import dataclasses
 from concurrent.futures import Future
 from contextlib import contextmanager, ExitStack
-import json
 import logging
-from datetime import datetime, timezone
+from datetime import datetime
 from queue import Queue, Empty
 import threading
-from typing import List, Optional, Mapping, Union, cast, Generator
+from time import time
+from typing import List, Optional, Union, cast, Generator, Dict
 from unittest import mock
 
 from google.api_core.exceptions import DeadlineExceeded
 from google.auth import environment_vars as google_env_vars, default as google_auth_default
 from google.cloud import pubsub_v1
-from google.cloud.pubsub_v1.proto.pubsub_pb2 import ReceivedMessage
-from google.cloud.pubsub_v1.subscriber.message import Message as PubSubMessage
-from google.cloud.pubsub_v1.types import FlowControl
+from google.cloud.pubsub_v1.subscriber.message import Message as SubscriberMessage
+from google.cloud.pubsub_v1.types import FlowControl, PubsubMessage, ReceivedMessage
+from google.protobuf.timestamp_pb2 import Timestamp
 
 from hedwig.backends.base import HedwigPublisherBaseBackend, HedwigConsumerBaseBackend
 from hedwig.backends.utils import override_env
@@ -63,74 +64,33 @@ def get_google_cloud_project() -> str:
     return settings.GOOGLE_CLOUD_PROJECT
 
 
-# TODO move to dataclasses in py3.7
+@dataclasses.dataclass(frozen=True)
 class GoogleMetadata:
     """
     Google Pub/Sub specific metadata for a Message
     """
 
-    def __init__(self, ack_id: str, subscription_path: str, publish_time: datetime, delivery_attempt: int):
-        self._ack_id: str = ack_id
-        self._subscription_path: str = subscription_path
-        self._publish_time: datetime = publish_time
-        self._delivery_attempt = delivery_attempt
+    ack_id: str
+    """
+    The ID used to ack the message
+    """
 
-    @property
-    def ack_id(self) -> str:
-        """
-        The ID used to ack the message
-        """
-        return self._ack_id
+    subscription_path: str
+    """
+    Path of the Pub/Sub subscription from which this message was pulled
+    """
 
-    @property
-    def subscription_path(self) -> str:
-        """
-        Path of the Pub/Sub subscription from which this message was pulled
-        """
-        return self._subscription_path
+    publish_time: datetime
+    """
+    Time this message was originally published to Pub/Sub
+    """
 
-    @property
-    def publish_time(self) -> datetime:
-        """
-        Time this message was originally published to Pub/Sub
-        """
-        return self._publish_time
-
-    @property
-    def delivery_attempt(self) -> int:
-        """
-        The delivery attempt counter received from Pub/Sub.
-        The first delivery of a given message will have this value as 1. The value
-        is calculated at best effort and is approximate.
-        """
-        return self._delivery_attempt
-
-    def __eq__(self, o: object) -> bool:
-        if not isinstance(o, GoogleMetadata):
-            return False
-
-        return (
-            self.ack_id == o.ack_id
-            and self.subscription_path == o.subscription_path  # noqa: W503
-            and self.delivery_attempt == o.delivery_attempt  # noqa: W503
-        )
-
-    def __ne__(self, o: object) -> bool:
-        return not self.__eq__(o)
-
-    def __str__(self) -> str:
-        return repr(self)
-
-    def __repr__(self) -> str:
-        return (
-            'GoogleMetadata('
-            f'ack_id={self.ack_id}, subscription_path={self.subscription_path}, publish_time={self.publish_time}, '
-            f'delivery_attempt={self.delivery_attempt}'
-            ')'
-        )
-
-    def __hash__(self) -> int:
-        return hash((self.ack_id, self.subscription_path, self.publish_time))
+    delivery_attempt: int
+    """
+    The delivery attempt counter received from Pub/Sub.
+    The first delivery of a given message will have this value as 1. The value
+    is calculated as best effort and is approximate.
+    """
 
 
 class GooglePubSubAsyncPublisherBackend(HedwigPublisherBaseBackend):
@@ -144,7 +104,7 @@ class GooglePubSubAsyncPublisherBackend(HedwigPublisherBaseBackend):
                 self._publisher = pubsub_v1.PublisherClient(batch_settings=settings.HEDWIG_PUBLISHER_GCP_BATCH_SETTINGS)
         return self._publisher
 
-    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Optional[Mapping] = None) -> Union[str, Future]:
+    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Dict[str, str]) -> Union[str, Future]:
         """
         Publishes to a Google Pub/Sub topic and returns a future that represents the publish API call. These API calls
         are batched for better performance.
@@ -152,38 +112,57 @@ class GooglePubSubAsyncPublisherBackend(HedwigPublisherBaseBackend):
         Note: despite the signature this doesn't return an actual instance of Future class, but an object that conforms
         to Future class. There's no generic type to represent future objects though.
         """
-        attrs = attrs or {}
         attrs = dict((str(key), str(value)) for key, value in attrs.items())
         return self.publisher.publish(topic_path, data=data, **attrs)
 
     def _get_topic_path(self, message: Message) -> str:
-        return self.publisher.topic_path(get_google_cloud_project(), f'hedwig-{message.topic}')
+        topic = self.topic(message)
+        if isinstance(topic, tuple):
+            topic, project = topic
+        else:
+            project = get_google_cloud_project()
+        return self.publisher.topic_path(project, f'hedwig-{topic}')
 
-    def _mock_queue_message(self, message: Message) -> mock.Mock:
-        gcp_message = mock.Mock()
-        gcp_message.message = mock.Mock()
-        gcp_message.message.data = json.dumps(message.as_dict()).encode('utf8')
-        gcp_message.message.publish_time = datetime.now(timezone.utc)
-        gcp_message.ack_id = 'test-receipt'
+    def _mock_queue_message(self, message: Message) -> "MessageWrapper":
+        payload, attributes = message.serialize()
+        # Pub/Sub requires bytes
+        if isinstance(payload, str):
+            payload = payload.encode('utf8')
+            attributes['hedwig_encoding'] = 'utf8'
+        publish_time = Timestamp()
+        publish_time.GetCurrentTime()
+        pubsub_message = PubsubMessage(
+            data=payload,
+            attributes=attributes,
+            message_id=str(int(time() * 1000)),
+            publish_time=publish_time,
+        )
+        # SubscriberMessage requires raw proto class, not proto-plus
+        subscriber_message = SubscriberMessage(PubsubMessage.pb(pubsub_message), 'test-receipt', 1, mock.MagicMock())
+        gcp_message = MessageWrapper(subscriber_message, 'test-subscription')
         return gcp_message
 
-    def _publish(self, message: Message, payload: str, headers: Optional[Mapping] = None) -> Union[str, Future]:
+    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> Union[str, Future]:
         topic_path = self._get_topic_path(message)
-        return self.publish_to_topic(topic_path, payload.encode('utf8'), headers)
+        # Pub/Sub requires bytes
+        if isinstance(payload, str):
+            payload = payload.encode('utf8')
+            attributes['hedwig_encoding'] = 'utf8'
+        return self.publish_to_topic(topic_path, payload, attributes)
 
 
 class GooglePubSubPublisherBackend(GooglePubSubAsyncPublisherBackend):
-    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Optional[Mapping] = None) -> Union[str, Future]:
+    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Dict[str, str]) -> Union[str, Future]:
         return cast(Future, super().publish_to_topic(topic_path, data, attrs)).result()
 
 
 class MessageWrapper:
-    def __init__(self, message: PubSubMessage, subscription_path: str):
+    def __init__(self, message: SubscriberMessage, subscription_path: str):
         self._message = message
         self._subscription_path = subscription_path
 
     @property
-    def message(self) -> PubSubMessage:
+    def message(self) -> SubscriberMessage:
         return self._message
 
     @property
@@ -207,13 +186,12 @@ class PubSubMessageScheduler:
         and the scheduling thread."""
         return self._queue
 
-    def schedule(self, callback, message: PubSubMessage) -> None:
+    def schedule(self, callback, message: SubscriberMessage) -> None:
         # callback is unused since we never set it in pull_messages
         self._work_queue.put(MessageWrapper(message, self._subscription_path))
 
     def shutdown(self) -> None:
-        """Shuts down the scheduler and immediately end all pending callbacks.
-        """
+        """Shuts down the scheduler and immediately end all pending callbacks."""
         # ideally we'd nack the messages in work queue, but that might take some time to finish.
         # instead, it's faster to actually process all the messages
 
@@ -232,10 +210,16 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
                     pubsub_v1.SubscriberClient.subscription_path(cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}-dlq')
                 ]
             else:
-                self._subscription_paths = [
-                    pubsub_v1.SubscriberClient.subscription_path(cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}-{x}')
-                    for x in settings.HEDWIG_SUBSCRIPTIONS
-                ]
+                self._subscription_paths = []
+                for subscription in settings.HEDWIG_SUBSCRIPTIONS:
+                    # all subscriptions live in an app's project, but cross-project subscriptions are named differently
+                    if isinstance(subscription, str):
+                        subscription_name = f'hedwig-{settings.HEDWIG_QUEUE}-{subscription}'
+                    else:
+                        subscription_name = f'hedwig-{settings.HEDWIG_QUEUE}-{subscription[1]}-{subscription[0]}'
+                    self._subscription_paths.append(
+                        pubsub_v1.SubscriberClient.subscription_path(cloud_project, subscription_name)
+                    )
                 # main queue for DLQ re-queued messages
                 self._subscription_paths.append(
                     pubsub_v1.SubscriberClient.subscription_path(cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}')
@@ -302,8 +286,14 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
             pass
 
     def process_message(self, queue_message: MessageWrapper) -> None:
+        # body is always bytes
+        message_payload = queue_message.message.data
+        attributes = queue_message.message.attributes
+        if attributes.get("hedwig_encoding") == "utf8":
+            message_payload = message_payload.decode('utf8')
         self.message_handler(
-            queue_message.message.data.decode(),
+            message_payload,
+            attributes,
             GoogleMetadata(
                 queue_message.message.ack_id,
                 queue_message.subscription_path,
@@ -333,7 +323,11 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
         """
         if visibility_timeout_s < 0 or visibility_timeout_s > 600:
             raise ValueError("Invalid visibility_timeout_s")
-        self.subscriber.modify_ack_deadline(metadata.subscription_path, [metadata.ack_id], visibility_timeout_s)
+        self.subscriber.modify_ack_deadline(
+            subscription=metadata.subscription_path,
+            ack_ids=[metadata.ack_id],
+            ack_deadline_seconds=visibility_timeout_s,
+        )
 
     def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: int = None) -> None:
         """
@@ -351,7 +345,10 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
         while True:
             try:
                 queue_messages: List[ReceivedMessage] = self.subscriber.pull(
-                    subscription_path, num_messages, retry=None, timeout=settings.GOOGLE_PUBSUB_READ_TIMEOUT_S
+                    subscription=subscription_path,
+                    max_messages=num_messages,
+                    retry=None,
+                    timeout=settings.GOOGLE_PUBSUB_READ_TIMEOUT_S,
                 ).received_messages
             except DeadlineExceeded:
                 break
@@ -364,7 +361,9 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
                 try:
                     if visibility_timeout:
                         self.subscriber.modify_ack_deadline(
-                            subscription_path, [queue_message.ack_id], visibility_timeout
+                            subscription=subscription_path,
+                            ack_ids=[queue_message.ack_id],
+                            ack_deadline_seconds=visibility_timeout,
                         )
 
                     future = self.publisher.publish(
@@ -377,7 +376,7 @@ class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
                         extra={'message_id': queue_message.message.message_id},
                     )
 
-                    self.subscriber.acknowledge(subscription_path, [queue_message.ack_id])
+                    self.subscriber.acknowledge(subscription=subscription_path, ack_ids=[queue_message.ack_id])
                 except Exception:
                     logger.exception('Exception in requeue message from {} to {}'.format(subscription_path, topic_path))
 
