@@ -3,7 +3,8 @@ import logging
 import os
 import importlib
 import typing
-from copy import deepcopy
+
+from hedwig.backends.import_utils import import_module_attr
 
 try:
     from django.conf import settings as django_settings
@@ -21,7 +22,6 @@ try:
 except ImportError:
     HAVE_FLASK = False
 
-
 _DEFAULTS: dict = {
     'AWS_REGION': None,
     'AWS_ACCOUNT_ID': None,
@@ -37,31 +37,31 @@ _DEFAULTS: dict = {
     'GOOGLE_PUBSUB_READ_TIMEOUT_S': 5,
     'HEDWIG_CALLBACKS': {},
     'HEDWIG_CONSUMER_BACKEND': None,
-    'HEDWIG_DATA_VALIDATOR_CLASS': 'hedwig.validator.MessageValidator',
+    'HEDWIG_DATA_VALIDATOR_CLASS': 'hedwig.validators.jsonschema.JSONSchemaValidator',
     'HEDWIG_DEFAULT_HEADERS': 'hedwig.conf.default_headers_hook',
     'HEDWIG_MESSAGE_ROUTING': {},
-    'HEDWIG_POST_DESERIALIZE_HOOK': 'hedwig.conf.noop_hook',
     'HEDWIG_PRE_PROCESS_HOOK': 'hedwig.conf.noop_hook',
     'HEDWIG_POST_PROCESS_HOOK': 'hedwig.conf.noop_hook',
-    'HEDWIG_PRE_SERIALIZE_HOOK': 'hedwig.conf.noop_hook',
     'HEDWIG_PUBLISHER': None,
     'HEDWIG_PUBLISHER_BACKEND': None,
     'HEDWIG_PUBLISHER_GCP_BATCH_SETTINGS': (),
     'HEDWIG_QUEUE': None,
-    'HEDWIG_SCHEMA_FILE': None,
+    'HEDWIG_JSONSCHEMA_FILE': None,
+    'HEDWIG_PROTOBUF_SCHEMA_MODULE': None,
     'HEDWIG_SYNC': False,
     'HEDWIG_SUBSCRIPTIONS': [],
+    'HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES': True,
 }
 
 
 # List of settings that may be in string import notation.
 _IMPORT_STRINGS = (
+    'HEDWIG_CONSUMER_BACKEND',
     'HEDWIG_DATA_VALIDATOR_CLASS',
     'HEDWIG_DEFAULT_HEADERS',
-    'HEDWIG_POST_DESERIALIZE_HOOK',
     'HEDWIG_PRE_PROCESS_HOOK',
     'HEDWIG_POST_PROCESS_HOOK',
-    'HEDWIG_PRE_SERIALIZE_HOOK',
+    'HEDWIG_PUBLISHER_BACKEND',
 )
 
 # List of settings that will be dicts with values as string import notation.
@@ -129,7 +129,7 @@ class _LazySettings:
         assert not self._user_settings, "Hedwig settings have already been configured"
 
         logging.info('Configuring Hedwig through object')
-        self._user_settings = deepcopy(obj)
+        self._user_settings = obj
 
     @staticmethod
     def _import_string(
@@ -142,17 +142,7 @@ class _LazySettings:
         if callable(dotted_path_or_callable):
             return dotted_path_or_callable
 
-        try:
-            module_path, class_name = dotted_path_or_callable.rsplit('.', 1)
-        except ValueError as err:
-            raise ImportError(f"{dotted_path_or_callable} doesn't look like a module path") from err
-
-        module = importlib.import_module(module_path)
-
-        try:
-            return getattr(module, class_name)
-        except AttributeError as err:
-            raise ImportError(f"Module '{module_path}' does not define a '{class_name}' attribute/class") from err
+        return import_module_attr(dotted_path_or_callable)
 
     def _get_setting_from_object(self, attr: str):
         if isinstance(self._user_settings, dict):
@@ -176,7 +166,7 @@ class _LazySettings:
         self._ensure_configured()
 
         if attr not in self._defaults:
-            raise AttributeError("Invalid API setting: '%s'" % attr)
+            raise AttributeError("Invalid setting: '%s'" % attr)
 
         try:
             val = self._get_setting_from_object(attr)
@@ -199,11 +189,23 @@ class _LazySettings:
         """
         Clear settings cache - useful for testing only
         """
+        from hedwig.backends.utils import get_publisher_backend, get_consumer_backend
+        from hedwig.callback import Callback
+        from hedwig.models import _validator
+
         for attr in self._defaults:
             try:
                 delattr(self, attr)
             except AttributeError:
                 pass
+        Callback.find_by_message.cache_clear()
+
+        # since consumer/publisher settings may have changed
+        get_publisher_backend.cache_clear()
+        get_consumer_backend.cache_clear()
+
+        # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
+        _validator.cache_clear()
 
 
 if HAVE_DJANGO:
@@ -225,5 +227,6 @@ This object allows settings to be accessed as properties. Settings can be config
 
 Some setting values need to be string import paths will be automatically resolved and return the class.
 
-Once settings have been configured, they can't be changed.
+Once settings have been configured, they shouldn't be changed. It is possible to re-configure for testing, but its not
+guaranteed to work the same way for non-test use cases.
 """
