@@ -5,11 +5,9 @@ from unittest import mock
 import pytest
 
 import hedwig.conf
-from hedwig import models
-from hedwig.backends.base import HedwigBaseBackend, HedwigPublisherBaseBackend
-from hedwig.backends.utils import get_publisher_backend, get_consumer_backend
-from hedwig.models import Message
-from hedwig.testing.factories import MessageFactory
+from hedwig.backends.base import HedwigPublisherBaseBackend
+from hedwig.backends.import_utils import import_module_attr
+from hedwig.testing.config import unconfigure
 
 from tests.models import MessageType
 
@@ -36,38 +34,48 @@ def settings():
         def __getattr__(self, name):
             return getattr(original_module, name)
 
+    unconfigure()
     hedwig.conf.settings._user_settings = Wrapped()
-    hedwig.conf.settings.clear_cache()
-
-    # since consumer/publisher settings may have changed
-    get_publisher_backend.cache_clear()
-    get_consumer_backend.cache_clear()
-
-    # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
-    models._validator = None
 
     try:
         yield hedwig.conf.settings._user_settings
     finally:
+        unconfigure()
         hedwig.conf.settings._user_settings = original_module
-        hedwig.conf.settings.clear_cache()
-
-        # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
-        models._validator = None
-
-        # since consumer/publisher settings may have changed
-        get_publisher_backend.cache_clear()
-        get_consumer_backend.cache_clear()
 
 
-@pytest.fixture(name='message_data')
-def _message_data():
-    return MessageFactory.build(msg_type=MessageType.trip_created)
+@pytest.fixture(name='message_factory', params=['jsonschema', 'protobuf'])
+def _message_factory(request, settings):
+    if request.param == 'jsonschema':
+        settings.HEDWIG_DATA_VALIDATOR_CLASS = 'hedwig.validators.jsonschema.JSONSchemaValidator'
+
+        try:
+            import jsonschema  # noqa
+            from hedwig.testing.factories.jsonschema import JSONSchemaMessageFactory  # noqa
+
+            yield JSONSchemaMessageFactory
+        except ImportError:
+            pytest.skip("JSON Schema not importable")
+
+    if request.param == 'protobuf':
+        settings.HEDWIG_DATA_VALIDATOR_CLASS = 'hedwig.validators.protobuf.ProtobufValidator'
+
+        try:
+            from tests.protobuf_factory import ProtobufMessageFactory  # noqa
+
+            yield ProtobufMessageFactory
+        except ImportError:
+            pytest.skip("Protobuf factory not importable")
 
 
 @pytest.fixture()
-def message(message_data):
-    return Message(message_data)
+def message_data(message_factory):
+    return message_factory.build(msg_type=MessageType.trip_created)
+
+
+@pytest.fixture()
+def message(message_factory):
+    return message_factory(msg_type=MessageType.trip_created)
 
 
 @contextmanager
@@ -101,21 +109,21 @@ def mock_pubsub_v1():
 def consumer_backend(request):
     if request.param == 'aws':
         try:
-            import hedwig.backends.aws  # noqa
+            from hedwig.backends.aws import AWSSQSConsumerBackend  # noqa
 
             with _mock_boto3():
-                yield HedwigBaseBackend.build("hedwig.backends.aws.AWSSQSConsumerBackend")
+                yield AWSSQSConsumerBackend()
         except ImportError:
             pytest.skip("AWS backend not importable")
 
     if request.param == 'google':
         try:
-            import hedwig.backends.gcp  # noqa
+            from hedwig.backends.gcp import GooglePubSubConsumerBackend  # noqa
 
             with mock.patch("hedwig.backends.gcp.pubsub_v1"), mock.patch(
                 "hedwig.backends.gcp.google_auth_default", return_value=(None, "DUMMY")
             ):
-                yield HedwigBaseBackend.build("hedwig.backends.gcp.GooglePubSubConsumerBackend")
+                yield GooglePubSubConsumerBackend()
         except ImportError:
             pytest.skip("Google backend not importable")
 
@@ -125,10 +133,16 @@ def consumer_backend(request):
 )
 def publisher_backend(request, mock_boto3):
     with mock.patch("hedwig.backends.gcp.pubsub_v1"):
-        yield HedwigBaseBackend.build(request.param)
+        yield import_module_attr(request.param)
 
 
 @pytest.fixture()
 def mock_publisher_backend():
     with mock.patch.object(HedwigPublisherBaseBackend, '_publish'):
         yield HedwigPublisherBaseBackend()
+
+
+@pytest.fixture(params=[True, False], ids=["message-attrs", "no-message-attrs"])
+def use_transport_message_attrs(request, settings):
+    settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES = request.param
+    yield settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES
